# -*- coding: utf-8 -*-
r"""
Module contains implementation of pyvista visualizations for spin lattices.
"""
from pathlib import Path

import vtkmodules.util.vtkConstants
from spinterface.visualizations.lattices.utilities import get_colormap, HSVtoRGB
import pyvista as pv
import vtk
import numpy as np
from spinterface.visualizations.lattices.ivisualizer import IVisualizer
from spinterface.inputs.lattice.ILattice import ILattice
from typing import List, Tuple, Union
from spinterface.visualizations.const import SPINDEFAULT_SETTINGS, EVECDEFAULT_SETTINGS, FORCEDEFAULT_SETTINGS
from spinterface.inputs.lattice.const import LATT_TYPE_EVEC, LATT_TYPE_SPIN, LATT_TYPE_FORCE


class CVisualPyVista(IVisualizer):
    r"""
    Class for visualizing spin lattices with py vista library
    """

    def __init__(self, lattice: ILattice, tiplength: Union[float, None] = None, tipradius: Union[float, None] = None,
                 arrowscale: Union[float, None] = None, draw_background: Union[bool, None] = None,
                 cam: Union[List[Tuple[float, float, float]], None] = None,
                 cmap: str = 'hsv_spind', total_mag: bool = False, topology: bool = False,
                 heatmap: bool = False, transparent_background: bool = True, color_background: str = "black",
                 heatmap_saturation: float = 0.35) -> None:
        r"""
        Initializes the visualization

        Args:
            tiplength(float): geometry of arrow: tiplength
            tipradius(float): geometry of arrow: tipradius
            arrowscale(float): geometry of arrow: arrowscale
            draw_background(bool): shall i draw the background of the lattice
            camera: camera position
            cmap: string for the choice of the colormap. Defined in utilities module
            total_mag(bool): if the sum over the lattice vectors shall be visualized in the center of the lattice
            topology(bool): if topology information shall be visualized
            heatmap(bool): the spin lattice can also be presented as a bivariate heat map
            heatmap_saturation(float): saturation; (radius of the hsv-color cylinder)
            transparent_background(bool): the background of the image
        """
        super().__init__(lattice)
        self.text_objects = []
        self.total_mag = total_mag
        self.topology = topology
        self.transparent_background = transparent_background
        self.color_background = color_background
        self.heatmap = heatmap
        self.heatmap_saturation = heatmap_saturation
        if heatmap and self.lattice.source == LATT_TYPE_EVEC:
            raise ValueError("Heatmap not coded yet for eigenvector visualization!")
        if heatmap and self.lattice.source == LATT_TYPE_FORCE:
            raise ValueError("Heatmap not coded yet for force visualization!")
        self.tiplength, self.tipradius, self.arrowscale, self.drawbackground = self._load_settings(tiplength, tipradius,
                                                                                                   arrowscale,
                                                                                                   draw_background)
        self._geom = pv.Arrow(start=np.array([-self.arrowscale / 2.0, 0, 0]), tip_length=self.tiplength,
                              tip_radius=self.tipradius, scale=self.arrowscale)
        self.cam = cam
        self.cmap = get_colormap(cmap)
        self._make_plotter()

    def _load_settings(self, tl: Union[float, None], tr: Union[float, None],
                       asc: Union[float, None], dbg: Union[bool, None]) -> Tuple[float, float, float, bool]:
        r"""
        Returns:
            loads the tiplength, tipradius and arrowscale depending on the inputs and the lattice type
        """
        # Decide on loading settings
        if self.lattice.source == LATT_TYPE_SPIN:
            print(f'loading defaults for type: {LATT_TYPE_SPIN}')
            tiplength = SPINDEFAULT_SETTINGS['tiplength']
            tipradius = SPINDEFAULT_SETTINGS['tipradius']
            arrowscale = SPINDEFAULT_SETTINGS['arrowscale']
            drawbackground = SPINDEFAULT_SETTINGS['drawbackground']
        elif self.lattice.source == LATT_TYPE_EVEC:
            print(f'loading defaults for type: {LATT_TYPE_EVEC}')
            tiplength = EVECDEFAULT_SETTINGS['tiplength']
            tipradius = EVECDEFAULT_SETTINGS['tipradius']
            arrowscale = EVECDEFAULT_SETTINGS['arrowscale']
            drawbackground = EVECDEFAULT_SETTINGS['drawbackground']
        elif self.lattice.source == LATT_TYPE_FORCE:
            print(f'loading defaults for type: {LATT_TYPE_FORCE}')
            tiplength = FORCEDEFAULT_SETTINGS['tiplength']
            tipradius = FORCEDEFAULT_SETTINGS['tipradius']
            arrowscale = FORCEDEFAULT_SETTINGS['arrowscale']
            drawbackground = FORCEDEFAULT_SETTINGS['drawbackground']
        else:
            raise ValueError('Not a valid lattice source!')
        if tl is not None:
            print('Overwriting tiplength setting with user input')
            tiplength = tl
        if tr is not None:
            print('Overwriting tiplradius setting with user input')
            tipradius = tr
        if asc is not None:
            print('Overwriting arrowscale setting with user input')
            arrowscale = asc
        if dbg is not None:
            print('Overwriting drawbackground setting with user input')
            drawbackground = dbg
        return tiplength, tipradius, arrowscale, drawbackground

    def add_meshes_to_different_plotter(self, plotter: pv.Plotter):
        r"""
        Adds the configuration and the meshes of this visualization to an external plotter
        """
        plotter.camera_position = self.cam
        if self.heatmap:
            plotpoints = self.lattice.points[:, :]
            self.PolyData = pv.PolyData(plotpoints)
            plotspins = self.lattice.spins[:, :]
            plot_colors = self._get_colors(plotspins=plotspins)
            self.PolyData['colors'] = plot_colors
            self.surf = self.PolyData.delaunay_2d()
            if self.topology:
                plotter.add_mesh(self.PolyData, scalars="colors", rgb=True, point_size=15,
                                      render_points_as_spheres=True)
            else:
                plotter.add_mesh(self.surf, scalars="colors", rgb=True)
        else:
            plotpoints, plotspins, plotsz = self._make_plot_points()
            self.PolyData = pv.PolyData(plotpoints)
            self.PolyData.vectors = plotspins
            self.PolyData['oop'] = plotsz
            if self.lattice.source == LATT_TYPE_SPIN:
                self.Glyphs = self.PolyData.glyph(orient=True, scale=True, geom=self._geom)
            elif self.lattice.source == LATT_TYPE_EVEC or self.lattice.source == LATT_TYPE_FORCE:
                self.Glyphs = self.PolyData.glyph(orient=True, scale=True, geom=self._geom)
            plotter.add_mesh(self.Glyphs, show_scalar_bar=False, cmap=self.cmap)
        for text_object in self.text_objects:
            self.plotter.add_text(**text_object)
        if self.total_mag:
            self._draw_total_magnetization()
        if self.topology:
            self._draw_topology()
        if self.drawbackground:
            self._draw_background()

    def _make_plotter(self, offscreen: bool = False):
        r"""
        Creates the plotter. The plotter will be recreated when saving the image
        """
        self.plotter = pv.Plotter(off_screen=offscreen, lighting='three lights')
        self._configureplotter()
        self.plotter.camera_position = self.cam
        if self.heatmap:
            plotpoints = self.lattice.points[:, :]
            self.PolyData = pv.PolyData(plotpoints)
            plotspins = self.lattice.spins[:, :]
            plot_colors = self._get_colors(plotspins=plotspins)
            self.PolyData['colors'] = plot_colors
            self.surf = self.PolyData.delaunay_2d()
            if self.topology:
                self.plotter.add_mesh(self.PolyData, scalars="colors", rgb=True, point_size=15,
                                      render_points_as_spheres=True)
            else:
                self.plotter.add_mesh(self.surf, scalars="colors", rgb=True)
        else:
            plotpoints, plotspins, plotsz = self._make_plot_points()
            self.PolyData = pv.PolyData(plotpoints)
            self.PolyData.vectors = plotspins
            self.PolyData['oop'] = plotsz
            if self.lattice.source == LATT_TYPE_SPIN:
                self.Glyphs = self.PolyData.glyph(orient=True, scale=True, geom=self._geom)
            elif self.lattice.source == LATT_TYPE_EVEC or self.lattice.source == LATT_TYPE_FORCE:
                self.Glyphs = self.PolyData.glyph(orient=True, scale=True, geom=self._geom)
            self.plotter.add_mesh(self.Glyphs, show_scalar_bar=False, cmap=self.cmap)
        if self.total_mag:
            self._draw_total_magnetization()
        if self.topology:
            self._draw_topology()
        if self.drawbackground:
            self._draw_background()

    def _get_colors(self, plotspins: np.ndarray) -> np.ndarray:
        r"""
        Assigns each spin a color calculate by bivariat HSV

        Args:
            plotspins(np.ndarray): array of spins
        Returns:
            2D np-array of rgb's
        """
        colors = []
        for spin in plotspins:
            colors.append(HSVtoRGB(spin, saturation=self.heatmap_saturation))
        return np.array(colors)

    def _draw_topology(self) -> None:
        r"""
        Visualizes the topology information of a lattice.
        """
        if self.lattice.source != LATT_TYPE_SPIN:
            print('WARNING: topology for non spin lattices might not make sense.')
        topos = self.lattice.topologies
        # draw the topologic density for each layer
        for layeridx in range(self.lattice.nlayer):
            topo = topos[layeridx]
            cell_types = np.array([vtk.VTK_TRIANGLE for count in range(len(topo.polygons1[::3, 1]))], np.int8)
            cells_list = []
            for idx in range(len(topo.polygons1[::3, 1])):
                cells_list.extend([3, idx * 3, idx * 3 + 1, idx * 3 + 2])
            cells = np.asarray(cells_list)
            grid1 = pv.UnstructuredGrid(cells, cell_types, topo.polygons1[:, :])
            grid1['dens'] = topo.areas1
            self.plotter.add_mesh(grid1, cmap='seismic')
            cell_types = np.array([vtk.VTK_TRIANGLE for count in range(len(topo.polygons2[::3, 1]))], np.int8)
            cells_list = []
            for idx in range(len(topo.polygons2[::3, 1])):
                cells_list.extend([3, idx * 3, idx * 3 + 1, idx * 3 + 2])
            cells = np.asarray(cells_list)
            grid2 = pv.UnstructuredGrid(cells, cell_types, topo.polygons2[:, :])
            grid2['dens'] = topo.areas2
            self.plotter.add_mesh(grid2, cmap='seismic')
            # visualize the center of mass with respect to topologic density.
            center = pv.PolyData(topo.topological_center)
            center = pv.Sphere(radius=1.0, center=topo.topological_center)
            # self.plotter.add_mesh(center, color='k')

    def _draw_total_magnetization(self) -> None:
        r"""
            the total magnetization (the sum of all spins) in the center of the lattice
        """
        length_tot_mag = np.linalg.norm(self.lattice.total_magnetization)

        if self.lattice.source == LATT_TYPE_EVEC:
            if length_tot_mag < 50:
                print('WARNING: the norm of the total magnetization is below 50. The visualization of it might not be '
                      'visible!')
            arrow = pv.Arrow(start=self.lattice.midpoint, direction=self.lattice.total_magnetization, tip_length=0.25,
                             tip_radius=0.1, tip_resolution=20,
                             shaft_radius=0.05, shaft_resolution=20, scale=length_tot_mag / 10)
        if self.lattice.source == LATT_TYPE_FORCE:
            if length_tot_mag < 50:
                print(
                    'WARNING: the norm of the total magnetization is below 50. The visualization of it might not be '
                    'visible!')
            arrow = pv.Arrow(start=self.lattice.midpoint, direction=self.lattice.total_magnetization,
                             tip_length=0.25,
                             tip_radius=0.1, tip_resolution=20,
                             shaft_radius=0.05, shaft_resolution=20, scale=length_tot_mag)
        else:
            if length_tot_mag < 500:
                print('WARNING: the norm of the total magnetization is below 500. The visualization of it might not be '
                      'visible!')
            arrow = pv.Arrow(start=self.lattice.midpoint, direction=self.lattice.total_magnetization, tip_length=0.25,
                             tip_radius=0.1, tip_resolution=20,
                             shaft_radius=0.05, shaft_resolution=20, scale=length_tot_mag / 100)
        self.plotter.add_mesh(arrow, color='k')

    def _draw_background(self) -> None:
        r"""
        Draws the background of the lattice
        """
        for layer in range(self.lattice.nlayer):
            magstruct = self.lattice.getlayer_by_idx(layer)
            points = magstruct[:, :3]
            points_poly = pv.PolyData(points)
            surface = points_poly.delaunay_2d()
            self.plotter.add_mesh(surface, show_edges=True, opacity=0.5)

    def _make_plot_points(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        r"""
        We always want to norm the colormap in the interval -1, 1 even we have a lattice which spins have only SZ comp.
        in the interval e.g. (1,0.5). There is now easy way to do this with pyvista since there is no interface for nor-
        malizing. Therefore, we add an invisible point in the center of the lattice here.

        Returns:
            the points, the spins and the sz components
        """
        plotpoints = np.append(self.lattice.points, np.array([self.lattice.midpoint, self.lattice.midpoint]), axis=0)
        plotspins = np.append(self.lattice.spins, np.array([[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]), axis=0)
        if self.lattice.source == LATT_TYPE_SPIN:
            plotsz = np.append(self.lattice.SZ, np.array([1.0, -1.0]))
        elif self.lattice.source == LATT_TYPE_EVEC or self.lattice.source == LATT_TYPE_FORCE:
            ez = np.array([0.0, 0.0, 1.0])
            plotsz = [np.dot(spin / np.linalg.norm(spin), ez) for spin in self.lattice.spins]
            plotsz = np.append(plotsz, np.array([1.0, -1.0]))
        else:
            raise ValueError('Lattice type not supported.')
        return plotpoints, plotspins, plotsz

    def _configureplotter(self) -> None:
        r"""
        Configures the plotter object
        """
        pv.set_plot_theme("ParaView")
        pv.rcParams['transparent_background'] = self.transparent_background
        self.plotter.set_background(self.color_background)

        def cam() -> None:
            print('Camera postion: ', self.plotter.camera_position)

        self.plotter.add_key_event('c', cam)

    def show(self) -> None:
        r"""
        Shows the plotter
        """
        print('Look what you have done.......')
        print('to get current cam-position press key c')
        self.plotter.show()

    def add_text(self, **kwargs):
        r"""
        """
        self.plotter.add_text(**kwargs)
        self.text_objects.append(dict(**kwargs))

    def __call__(self, outpath: Path = Path.cwd() / 'spin.png') -> None:
        r"""
        Saves the image to a file

        Args:
            outpath(Path): output path for the png image created.
        """
        self._make_plotter(offscreen=True)
        for text_object in self.text_objects:
            self.plotter.add_text(**text_object)
        self.plotter.window_size = [4000, 4000]
        self.plotter.screenshot(str(outpath))



import numpy as np
import pandas as pd
import pyvista as pv
from pathlib import Path
from spinterface.visualizations.energylandscapes.i2dfunction import I2DFunction
from spinterface.visualizations.energylandscapes.cvisualpyvista import CVisualPyVista
from spinterface.algorithms.spineapple.minimizer.cnewton_2d import CNewton2D
from spinterface.algorithms.spineapple.gneb.cneb2d import CNEB2D,CciNEB2d


class Clandscape(I2DFunction):
    r"""

    """

    def __call__(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        omega_x = 1 / 4 * np.pi  # should produce minimum for x=+-4
        omega_y = 1 / 3 * np.pi  # should produce minimum for y=+-3
        return 1.0 * np.cos(1.0 * (omega_x * x)) + np.cos((omega_y * y)) + \
               self.gauss(x, y, x0=4, y0=3, a=4.0, sigmax=5, sigmay=5) + \
               self.gauss(x, y, x0=-4, y0=-3, a=4.0, sigmax=5, sigmay=5) + \
               self.gauss(x, y, x0=0, y0=-8, a=2.0, sigmax=5, sigmay=5) + \
               self.gauss(x, y, x0=1.5, y0=-3, a=1.5, sigmax=1, sigmay=2) + \
               self.gauss(x, y, x0=4, y0=-6, a=3.0, sigmax=3, sigmay=3) + \
               self.gauss(x, y, x0=4.0, y0=4.0, a=2.0, sigmax=3.2, sigmay=3.2)


x = np.linspace(-6, 5.0, 500)
y = np.linspace(-6, 4, 500)

func = Clandscape()
visualizer = CVisualPyVista(x=x, y=y, function=func, cam=[(9.094746928770046, 16.863633966086084, 29.17188311757661),
 (0.0, 0.0, -0.1160760797235747),
 (-0.35791776521192226, -0.7568231363882767, 0.5469128025317338)])


image = 5
point_size_big = 80
point_size_small = 50
# guess points near the minimum based on a view from the top===================================
#visualizer.show_grid_topview()

# start point to minimize into the a state:
newton_minimizer = CNewton2D(f=func, initial=np.array([2, -4]))
Astate = newton_minimizer()
if image in [1]:
    visualizer.placepointonsurface(xypoint=np.array([2, -4]), point_size=point_size_big, color='k')
    A_initial_projected = visualizer.projectosurface(np.array([2,-4]))
    A_projected=visualizer.projectosurface(Astate[-1])
    minimization_vector = pv.Arrow(start=A_initial_projected,direction=A_projected-A_initial_projected)
    visualizer.add_mesh(mesh=minimization_vector, color='k')
if image in [1,2,3,4,5]:
    visualizer.placepointonsurface(xypoint=Astate[-1], color='red', point_size=point_size_big)
#.............................................................

# start point to minimize into the m state
newton_minimizer = CNewton2D(f=func, initial=np.array([-1, -4]))
Mstate = newton_minimizer()
if image in [4]:
    M_initial_projected = visualizer.projectosurface(np.array([0,-1.8]))
    M_projected=visualizer.projectosurface(Mstate[-1])
    minimization_vector = pv.Arrow(start=M_initial_projected,direction=M_projected-M_initial_projected)
    visualizer.add_mesh(minimization_vector, color='k')
if image in [4,5]:
    visualizer.placepointonsurface(xypoint=Mstate[-1], color='red', point_size=point_size_big)

#visualizer.placepointonsurface(xypoint=Mstate[-1], color='red')
#.............................................................

# start point to minimize into the b state
newton_minimizer = CNewton2D(f=func, initial=np.array([-4, 2]))
Bstate = newton_minimizer()
if image in [1]:
    visualizer.placepointonsurface(xypoint=np.array([-4, 2]), point_size=point_size_big, color='k')
    B_initial_projected = visualizer.projectosurface(np.array([-4,2]))
    B_projected=visualizer.projectosurface(Bstate[-1])
    minimization_vector = pv.Arrow(start=B_initial_projected,direction=B_projected-B_initial_projected)
    visualizer.add_mesh(minimization_vector, color='k')
if image in [1,2,3,4,5]:
    visualizer.placepointonsurface(xypoint=Bstate[-1], color='red', point_size=point_size_big)
#.............................................................

# direct connection between A and B:
initial_path = visualizer.create_sequence_between_points(initial=Astate[-1], final=Bstate[-1], nr=15)
if image in [2]:
    visualizer.placepointsequenceonsurface(initial_path, color_points='gray', point_size=point_size_small, tuberadius=0.05)
#.............................................................

# constructed result after short NEB calculations
cut_point = np.array([0,-1.8])
initial_path_AM = visualizer.create_sequence_between_points(initial=Astate[-1], final=cut_point, nr=10)
initial_path_MB = visualizer.create_sequence_between_points(initial=cut_point, final=Bstate[-1], nr=10)
if image in [3,4]:
    if image in [3]:
        visualizer.placepointsequenceonsurface(initial_path_AM, color_points='gray', point_size=point_size_small, tuberadius=0.05)
        visualizer.placepointsequenceonsurface(initial_path_MB, color_points='gray', point_size=point_size_small, tuberadius=0.05)
    visualizer.placepointonsurface(xypoint=cut_point, point_size=point_size_big, color='k')
#.............................................................

# GNEB A to M
nebber = CNEB2D(f=func,initial=Astate[-1],final=Mstate[-1],nim=12)
#path_AM = nebber(i_write_out=True, write_out=Path.cwd() / 'path_AM.dat')
path_AM = nebber.read_convergedpath_fromfile(Path.cwd() / 'path_AM.dat')
ci_nebber = CciNEB2d(function=func, file=Path.cwd() / 'path_AM.dat')
#path_AMsp, sp = ci_nebber(i_write_out=True,write_out=Path.cwd() / 'path_AM_ci.dat')
path_AMsp = ci_nebber.read_convergedpath_fromfile(Path.cwd() / 'path_AM_ci.dat')
ci = ci_nebber.find_ci()

if image in [5]:
    visualizer.placepointsequenceonsurface(path_AMsp, color_points='blue', point_size=point_size_small,tuberadius=0.05)
    visualizer.placepointonsurface(xypoint=path_AMsp[ci], color='green', point_size=point_size_big)
#.............................................................

# GNEB M to B
nebber = CNEB2D(f=func,initial=Mstate[-1],final=Bstate[-1],nim=12)
#path_MB = nebber(i_write_out=True, write_out=Path.cwd() / 'path_MB.dat')
path_MB = nebber.read_convergedpath_fromfile(Path.cwd() / 'path_MB.dat')
ci_nebber = CciNEB2d(function=func, file=Path.cwd() / 'path_MB.dat')
#path_MBsp, sp = ci_nebber(i_write_out=True,write_out=Path.cwd() / 'path_MB_ci.dat')
path_MBsp = ci_nebber.read_convergedpath_fromfile(Path.cwd() / 'path_MB_ci.dat')
ci = ci_nebber.find_ci()
if image in [5]:
    visualizer.placepointsequenceonsurface(path_MBsp, color_points='blue', point_size=point_size_small,tuberadius=0.05)
    visualizer.placepointonsurface(xypoint=path_MBsp[ci], color='green', point_size=point_size_big)
#.............................................................

visualizer.show()
visualizer(outpath=Path.cwd() / f'method{image}.png')



r"""
Combines different visualization in a grid of plots
"""
from spinterface.visualizations.lattices.cvisualpyvista import CVisualPyVista
from typing import List, Tuple
import pyvista as pv
from pathlib import Path


class CVisualCollection:
    r"""

    """

    def __init__(self, frames: List[CVisualPyVista], grid: Tuple[int, int], windowsize: Tuple[float,float] = (4000,4000)) -> None:
        r"""

        :param frames:
        :param grid:
        """
        self.grid = grid
        self.frames = frames
        self.windowsize = windowsize
        nr_frames = self.grid[0] * self.grid[1]
        if len(frames) > nr_frames:
            print('WARNING: number of images exceeds grid. Only the first images will be presented...')
        if len(frames) < nr_frames:
            print('WARNING: number of images less then possible for this grid. There will be free elements...')
        self._make_multiplot()

    def _make_multiplot(self, offscreen: bool = False):
        r"""
        Makes the multiplot
        """
        if offscreen:
            self.plotter = pv.Plotter(shape=self.grid, off_screen=offscreen, lighting='three lights', window_size=self.windowsize)
        else:
            self.plotter = pv.Plotter(shape=self.grid, off_screen=offscreen, lighting='three lights')
        ind1 = 0
        ind2 = 0
        for frame in self.frames:
            if ind1 == self.grid[0]:
                ind2 = ind2 + 1
                ind1 = 0
            self.plotter.subplot(ind1, ind2)
            frame.add_meshes_to_different_plotter(self.plotter)
            ind1 = ind1 + 1

    def show(self) -> None:
        r"""
        Shows the plotter
        """
        print('Look what you have done.......')
        self.plotter.show()

    def __call__(self, outpath: Path = Path.cwd() / 'collection.png') -> None:
        r"""
        Saves the image to a file

        Args:
            outpath(Path): output path for the png image created.
        """
        self._make_multiplot(offscreen=True)
        self.plotter.screenshot(str(outpath))



# -*- coding: utf-8 -*-
r"""
Python implementation of the NEB-method for a two dimensional configuration space. Greatly inspired by the
implementation of Stephan von Malottki. This can be used to create visualization of the paths and energy landscapes for
low dimensional systems. Useful for shematic explanations of the GNEB method in presentations or texts.
"""
import numpy as np
from typing import Callable, Union, Tuple
from pathlib import Path
import pandas as pd


class CNEB2D:
    r"""
    NEB-method for 2D configuration space
    """

    def __init__(self, f: Callable, initial: np.ndarray, final: np.ndarray, nim: int, ds: float = 0.00005,
                 kappa: float = 0.5, converge: float = 0.05, read_path_from_file: Union[Path, None] = None) -> None:
        r"""
        Initializer for the NEB-method

        Args:
            f(Callable): z = f(x,y)

            initial(np.ndarray): initial point for the path

            final(np.ndarray): final point for the path

            nim(int): number of images

            ds(float):

            kappa(float):

            converge(float):

            read_path_from_file: Read path from file. Default is None. In that case an initial path is created as the
            shortest distance between initial and final (line).

        """
        # physical inputs
        self._f = self._function_wrapper(f)
        if (len(initial) != 2) or (len(final) != 2):
            raise ValueError('Initial and final states have to be 2D points on the XY-plane.')
        self._initial = initial
        self._final = final
        self._nim = nim
        # simulation parameters:
        self._ds = ds
        self._kappa = kappa
        self._converge = converge
        self._displacement = converge * 5
        self._displacement_old = converge * 15.0
        # outputs:
        if read_path_from_file is None:
            self._path = self._create_initial_path()
        else:
            self._path = self.read_convergedpath_fromfile(file=read_path_from_file)

    def _function_wrapper(self, f: Callable) -> Callable:
        r"""
        This wrapper changes the input structure of the callable to be in line with the visualization interface. The
        visualization interface uses a function which takes f(x,y) while the methods in this class need a callable with
        the parameter f(r(x,y)). This wrapper provides this functionality.
        Args:
            f: takes a callable with the functionality f(x,y)

        Returns: a function with f(r).
        """

        def wrapper(r: np.ndarray) -> np.ndarray:
            if len(np.shape(r)) == 1:
                x = r[0]
                y = r[1]
            else:
                x = r[:, 0]
                y = r[:, 1]
            return f(x, y)

        return wrapper

    def _create_initial_path(self) -> np.ndarray:
        r"""
        Creates an initial path as the linear connection between the initial and final state.

        Returns:
            the initial path: [[x_ini,y_ini],[x1,y1],[x2,y2],...,[x_fin,y_fin]]
        """
        dx = (self._final[0] - self._initial[0]) / (self._nim - 1)
        dy = (self._final[1] - self._initial[1]) / (self._nim - 1)
        return np.array([[self._initial[0] + dx * i, self._initial[1] + dy * i] for i in range(self._nim)])

    def __call__(self, i_write_out: bool = False, gralog: int = 1000,
                 write_out: Path = Path.cwd() / 'NEB.dat') -> np.ndarray:
        r"""
        Calls the NEB algorithm!

        Args:
            gralog(int): frequency for writing the output
            write_out(True): in the frequency of gralog write the actual path to a file.

        Returns:
            the converged path
        """
        i = 1
        if i_write_out:
            write_dict = {'iteration': [0], 'force': [self._displacement], 'path': [self._path]}

        while self._displacement > self._converge:
            if np.mod(i, gralog) == 0:
                print(f"iteration GNEB: {i}, {self._displacement}")
                # Here we can produce the outputs!
                if i_write_out:
                    write_dict['iteration'].append(i)
                    write_dict['force'].append(self._displacement)
                    write_dict['path'].append(self._path)
                    df = pd.DataFrame(write_dict)
                    df.to_json(write_out)
            # updates self._path and self._displacement!
            self._relax_path_NEB()
            i += 1

        # final
        if i_write_out:
            write_dict['iteration'].append(i)
            write_dict['force'].append(self._displacement)
            write_dict['path'].append(self._path)
            df = pd.DataFrame(write_dict)
            df.to_json(write_out)
        return self._path

    def _relax_path_NEB(self):
        r"""
        Does one relaxations step with the forces of NEB algorithm
        """
        path_new = self._path.copy()
        l_dist, l_sd_r, l_sd_l = self._springs()
        force = 0.0

        for idx in range(1, len(self._path) - 1):
            dp = 0.01
            p = self._path[idx]
            dfdx, dfdy = self._calc_derivative(p, dp)
            dEr = self._f(self._path[idx + 1]) - self._f(p)
            dEl = self._f(p) - self._f(self._path[idx - 1])
            dE_max = max(np.abs(dEr), np.abs(dEl))
            dE_min = min(np.abs(dEr), np.abs(dEl))
            sd = l_sd_r[idx]

            if self._f(self._path[idx + 1]) > self._f(p) > self._f(self._path[idx - 1]):
                sd = l_sd_r[idx]
            elif self._f(self._path[idx + 1]) < self._f(p) < self._f(self._path[idx - 1]):
                sd = l_sd_l[idx]
            elif self._f(self._path[idx + 1]) > self._f(self._path[idx - 1]):
                sd[0] = l_sd_r[idx][0] * dE_max + l_sd_l[idx][0] * dE_min
                sd[1] = l_sd_r[idx][1] * dE_max + l_sd_l[idx][1] * dE_min
            else:
                sd[0] = l_sd_r[idx][0] * dE_min + l_sd_l[idx][0] * dE_max
                sd[1] = l_sd_r[idx][1] * dE_min + l_sd_l[idx][1] * dE_max
            # normalize direction
            sd = sd / np.linalg.norm(sd)
            dfdx_new = dfdx - np.dot([dfdx, dfdy], sd) * sd[0]
            dfdy_new = dfdy - np.dot([dfdx, dfdy], sd) * sd[1]
            dfdx = dfdx_new
            dfdy = dfdy_new
            force = max(force, np.sqrt(
                (-dfdx + sd[0] * l_dist[idx] * self._kappa) ** 2 + (-dfdy + sd[1] * l_dist[idx] * self._kappa) ** 2))

        for i in range(1, len(self._path) - 1):
            dp = 0.01
            case = 0
            p = self._path[i]
            dfdx, dfdy = self._calc_derivative(p, dp)
            dEr = self._f(self._path[i + 1]) - self._f(self._path[i])
            dEl = self._f(self._path[i]) - self._f(self._path[i - 1])
            dE_max = max(np.abs(dEr), np.abs(dEl))
            dE_min = min(np.abs(dEr), np.abs(dEl))
            sd = l_sd_r[i]
            if self._f(self._path[i + 1]) > self._f(self._path[i]) > self._f(self._path[i - 1]):
                sd = l_sd_r[i]
                case = 1
            elif self._f(self._path[i + 1]) < self._f(self._path[i]) < self._f(self._path[i - 1]):
                sd = l_sd_l[i]
                case = 2
            elif self._f(self._path[i + 1]) > self._f(self._path[i - 1]):
                sd[0] = l_sd_r[i][0] * dE_max + l_sd_l[i][0] * dE_min
                sd[1] = l_sd_r[i][1] * dE_max + l_sd_l[i][1] * dE_min
                case = 3
            else:
                sd[0] = l_sd_r[i][0] * dE_min + l_sd_l[i][0] * dE_max
                sd[1] = l_sd_r[i][1] * dE_min + l_sd_l[i][1] * dE_max
                case = 4
            sd = sd / np.linalg.norm(sd)
            dfdx_new = dfdx - np.dot([dfdx, dfdy], sd) * sd[0]
            dfdy_new = dfdy - np.dot([dfdx, dfdy], sd) * sd[1]
            dfdx = dfdx_new
            dfdy = dfdy_new
            image_new_x = self._path[i][0] + (-dfdx + sd[0] * l_dist[i] * self._kappa) * self._ds / force
            image_new_y = self._path[i][1] + (-dfdy + sd[1] * l_dist[i] * self._kappa) * self._ds / force
            path_new[i][0] = image_new_x
            path_new[i][1] = image_new_y
        self._displacement = force
        self._path = path_new

    def _springs_previous(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        r"""

        """
        sd_r = np.zeros(shape=(len(self._path) - 1, 2))
        sd_l = np.zeros(shape=(len(self._path) - 1, 2))
        sd_r[1:] = np.diff(self._path[:-1], axis=0)
        sd_l[1:] = np.diff(self._path[1:], axis=0)
        distance = np.linalg.norm(sd_r, axis=1) - np.linalg.norm(sd_l, axis=1)
        return distance, sd_r, sd_l

    def _springs(self):
        distance = [0]
        sd_r = [[0, 0]]
        sd_l = [[0, 0]]
        for i in range(1, len(self._path) - 1):
            dx_l = self._path[i][0] - self._path[i - 1][0]
            dx_r = self._path[i + 1][0] - self._path[i][0]
            dy_l = self._path[i][1] - self._path[i - 1][1]
            dy_r = self._path[i + 1][1] - self._path[i][1]
            dr_l = np.sqrt(dx_l ** 2 + dy_l ** 2)
            dr_r = np.sqrt(dx_r ** 2 + dy_r ** 2)
            ddr = dr_r - dr_l
            distance.append(ddr)
            #      spring_directions.append([(dx_r/norm_r+dx_l/norm_l)/2.0,(dy_r/norm_r+dy_l/norm_l)/2.0])
            #      norm_sd=np.sqrt((dx_r+dx_l)**2+(dy_r+dy_l)**2)
            #      sd_x=(dx_r+dx_l)/norm_sd
            #      sd_y=(dy_r+dy_l)/norm_sd
            sd_r.append([dx_r, dy_r])
            sd_l.append([dx_l, dy_l])
        return distance, sd_r, sd_l

    def _calc_derivative(self, p: np.ndarray, dp: float) -> Tuple[float, float]:
        r"""
        Calculates the finite element derivative of the function f for the image p

        Args:
            p(np.ndarray): current image
            dp(float): step width

        Returns:
            dfdx: derivative in x direction, dfdy: derivative in y direction
        """
        dfdx = (self._f([p[0] + dp, p[1]]) - self._f([p[0] - dp, p[1]])) / (2.0 * dp)
        dfdy = (self._f([p[0], p[1] + dp]) - self._f([p[0], p[1] - dp])) / (2.0 * dp)
        return dfdx, dfdy

    @staticmethod
    def read_convergedpath_fromfile(file: Path) -> np.ndarray:
        r"""
        The path is sometimes written to files to avoid re-computing the path again and again

        Args:
            file: path saved in file

        Returns:
            the last path (mostly the converged one) (sequence of points)
        """
        df = pd.read_json(file)
        max_iter = np.max(df['iteration'].to_numpy())
        path = df['path'][df['iteration'] == max_iter].to_list()
        pl = []
        for p in path[0]:
            pl.append(np.array(p))
        return np.vstack(tuple(pl))


class CciNEB2d:
    r"""
    The climbing image implementation of the two dimensional NEB. Inheriting from the NEB method.
    """

    def __init__(self, function: Callable, file: Path, ds: float = 0.00005, kappa: float = 0.5,
                 converge: float = 0.05) -> None:
        r"""
        Initializes the ci GNEB method

        Args:
            path: file of a converged GNEB calculation. The climbing image method needs some kind of start path
        """
        self._path = self.read_convergedpath_fromfile(file)
        self._f = self._function_wrapper(function)
        self._kappa = kappa
        self._converge = converge
        self._ds = ds
        self._displacement = converge * 5
        self._ci = self.find_ci()

    def find_ci(self) -> int:
        r"""
        Returns:
            the index of the current climbing image (image with the highest energy
        """
        ci = 0
        z_max = self._f(self._path[0])
        for (i, image) in enumerate(self._path):
            if self._f(image) > z_max:
                z_max = self._f(image)
                ci = i
        return ci

    def _function_wrapper(self, f: Callable) -> Callable:
        r"""
        This wrapper changes the input structure of the callable to be in line with the visualization interface. The
        visualization interface uses a function which takes f(x,y) while the methods in this class need a callable with
        the parameter f(r(x,y)). This wrapper provides this functionality.
        Args:
            f: takes a callable with the functionality f(x,y)

        Returns: a function with f(r).
        """

        def wrapper(r: np.ndarray) -> np.ndarray:
            if len(np.shape(r)) == 1:
                x = r[0]
                y = r[1]
            else:
                x = r[:, 0]
                y = r[:, 1]
            return f(x, y)

        return wrapper

    def __call__(self, i_write_out: bool = False, gralog: int = 1000,
                 write_out: Path = Path.cwd() / 'ciNEB.dat') -> Tuple[np.ndarray, int]:
        r"""
        Calls the ciNEB algorithm!

        Args:
            gralog(int): frequency for writing the output
            write_out(True): in the frequency of gralog write the actual path to a file.

        Returns:
            the converged path, and the saddlepoint
        """
        i = 1
        if i_write_out:
            write_dict = {'iteration': [0], 'force': [self._displacement], 'path': [self._path]}

        while self._displacement > self._converge:
            if np.mod(i, gralog) == 0:
                print(f"iteration ciNEB: {i}, {self._displacement}")
                # Here we can produce the outputs!
                if i_write_out:
                    write_dict['iteration'].append(i)
                    write_dict['force'].append(self._displacement)
                    write_dict['path'].append(self._path)
                    df = pd.DataFrame(write_dict)
                    df.to_json(write_out)
            # updates self._path and self._displacement!
            self._ci = self.find_ci()
            l_displacement_old = self._displacement
            self._relax_path_ciNEB()
            if abs(l_displacement_old-self._displacement) <= 0.0000001:
                break
            i += 1

        # final
        if i_write_out:
            write_dict['iteration'].append(i)
            write_dict['force'].append(self._displacement)
            write_dict['path'].append(self._path)
            df = pd.DataFrame(write_dict)
            df.to_json(write_out)
        return self._path, self._ci

    def _relax_path_ciNEB(self):
        r"""
        Does one relaxations step with the forces of ciNEB algorithm
        """
        path_new = self._path.copy()
        l_dist, l_sd_r, l_sd_l = self._springs()
        force = 0.0

        for idx in range(1, len(self._path) - 1):
            dp = 0.01
            p = self._path[idx]
            dfdx, dfdy = self._calc_derivative(p, dp)
            dEr = self._f(self._path[idx + 1]) - self._f(p)
            dEl = self._f(p) - self._f(self._path[idx - 1])
            dE_max = max(np.abs(dEr), np.abs(dEl))
            dE_min = min(np.abs(dEr), np.abs(dEl))
            sd = l_sd_r[idx]

            if self._f(self._path[idx + 1]) > self._f(p) > self._f(self._path[idx - 1]):
                sd = l_sd_r[idx]
            elif self._f(self._path[idx + 1]) < self._f(p) < self._f(self._path[idx - 1]):
                sd = l_sd_l[idx]
            elif self._f(self._path[idx + 1]) > self._f(self._path[idx - 1]):
                sd[0] = l_sd_r[idx][0] * dE_max + l_sd_l[idx][0] * dE_min
                sd[1] = l_sd_r[idx][1] * dE_max + l_sd_l[idx][1] * dE_min
            else:
                sd[0] = l_sd_r[idx][0] * dE_min + l_sd_l[idx][0] * dE_max
                sd[1] = l_sd_r[idx][1] * dE_min + l_sd_l[idx][1] * dE_max
            # normalize direction
            sd = sd / np.linalg.norm(sd)
            if idx == self._ci:
                dfdx_new = dfdx - 2.0 * np.dot([dfdx, dfdy], sd) * sd[0]
                dfdy_new = dfdy - 2.0 * np.dot([dfdx, dfdy], sd) * sd[1]
                dfdx = dfdx_new
                dfdy = dfdy_new
            else:
                dfdx_new = dfdx - np.dot([dfdx, dfdy], sd) * sd[0]
                dfdy_new = dfdy - np.dot([dfdx, dfdy], sd) * sd[1]
                dfdx = dfdx_new
                dfdy = dfdy_new
            force = max(force, np.sqrt(
                (-dfdx + sd[0] * l_dist[idx] * self._kappa) ** 2 + (-dfdy + sd[1] * l_dist[idx] * self._kappa) ** 2))

        for i in range(1, len(self._path) - 1):
            dp = 0.01
            case = 0
            p = self._path[i]
            dfdx, dfdy = self._calc_derivative(p, dp)
            dEr = self._f(self._path[i + 1]) - self._f(self._path[i])
            dEl = self._f(self._path[i]) - self._f(self._path[i - 1])
            dE_max = max(np.abs(dEr), np.abs(dEl))
            dE_min = min(np.abs(dEr), np.abs(dEl))
            sd = l_sd_r[i]
            if self._f(self._path[i + 1]) > self._f(self._path[i]) > self._f(self._path[i - 1]):
                sd = l_sd_r[i]
                case = 1
            elif self._f(self._path[i + 1]) < self._f(self._path[i]) < self._f(self._path[i - 1]):
                sd = l_sd_l[i]
                case = 2
            elif self._f(self._path[i + 1]) > self._f(self._path[i - 1]):
                sd[0] = l_sd_r[i][0] * dE_max + l_sd_l[i][0] * dE_min
                sd[1] = l_sd_r[i][1] * dE_max + l_sd_l[i][1] * dE_min
                case = 3
            else:
                sd[0] = l_sd_r[i][0] * dE_min + l_sd_l[i][0] * dE_max
                sd[1] = l_sd_r[i][1] * dE_min + l_sd_l[i][1] * dE_max
                case = 4
            sd = sd / np.linalg.norm(sd)
            if idx == self._ci:
                dfdx_new = dfdx - 2.0 * np.dot([dfdx, dfdy], sd) * sd[0]
                dfdy_new = dfdy - 2.0 * np.dot([dfdx, dfdy], sd) * sd[1]
                dfdx = dfdx_new
                dfdy = dfdy_new
            else:
                dfdx_new = dfdx - np.dot([dfdx, dfdy], sd) * sd[0]
                dfdy_new = dfdy - np.dot([dfdx, dfdy], sd) * sd[1]
                dfdx = dfdx_new
                dfdy = dfdy_new
            if idx == self._ci:
                image_new_x = self._path[i][0] + (-dfdx) * self._ds / force
                image_new_y = self._path[i][1] + (-dfdy) * self._ds / force
            else:
                image_new_x = self._path[i][0] + (-dfdx + sd[0] * l_dist[i] * self._kappa) * self._ds / force
                image_new_y = self._path[i][1] + (-dfdy + sd[1] * l_dist[i] * self._kappa) * self._ds / force
            path_new[i][0] = image_new_x
            path_new[i][1] = image_new_y
        self._displacement = force
        self._path = path_new

    @staticmethod
    def read_convergedpath_fromfile(file: Path) -> np.ndarray:
        r"""
        The path is sometimes written to files to avoid re-computing the path again and again

        Args:
            file: path saved in file

        Returns:
            the last path (mostly the converged one) (sequence of points)
        """
        df = pd.read_json(file)
        max_iter = np.max(df['iteration'].to_numpy())
        path = df['path'][df['iteration'] == max_iter].to_list()
        pl = []
        for p in path[0]:
            pl.append(np.array(p))
        return np.vstack(tuple(pl))

    def _springs(self):
        distance = [0]
        sd_r = [[0, 0]]
        sd_l = [[0, 0]]
        for i in range(1, len(self._path) - 1):
            dx_l = self._path[i][0] - self._path[i - 1][0]
            dx_r = self._path[i + 1][0] - self._path[i][0]
            dy_l = self._path[i][1] - self._path[i - 1][1]
            dy_r = self._path[i + 1][1] - self._path[i][1]
            dr_l = np.sqrt(dx_l ** 2 + dy_l ** 2)
            dr_r = np.sqrt(dx_r ** 2 + dy_r ** 2)
            ddr = dr_r - dr_l
            distance.append(ddr)
            #      spring_directions.append([(dx_r/norm_r+dx_l/norm_l)/2.0,(dy_r/norm_r+dy_l/norm_l)/2.0])
            #      norm_sd=np.sqrt((dx_r+dx_l)**2+(dy_r+dy_l)**2)
            #      sd_x=(dx_r+dx_l)/norm_sd
            #      sd_y=(dy_r+dy_l)/norm_sd
            sd_r.append([dx_r, dy_r])
            sd_l.append([dx_l, dy_l])
        return distance, sd_r, sd_l

    def _calc_derivative(self, p: np.ndarray, dp: float) -> Tuple[float, float]:
        r"""
        Calculates the finite element derivative of the function f for the image p

        Args:
            p(np.ndarray): current image
            dp(float): step width

        Returns:
            dfdx: derivative in x direction, dfdy: derivative in y direction
        """
        dfdx = (self._f([p[0] + dp, p[1]]) - self._f([p[0] - dp, p[1]])) / (2.0 * dp)
        dfdy = (self._f([p[0], p[1] + dp]) - self._f([p[0], p[1] - dp])) / (2.0 * dp)
        return dfdx, dfdy



r"""
Constants for setting up the main CSPS class. Contains:
- Necessary keys and their data types in settings
"""
# Necessary keys =======================================================================================================
KEYS = {'sps_attempts': int, 'test_convex_initial': bool, 'algo_displace': int, 'i_cycle_after_displace': bool,
        'disp2_superpos_modenr_low': int, 'disp2_superpos_modenr_high': int, 'disp2_eigalgo': int,
        'disp2_stepscale': float,
        'algo_escape': int, 'i_start_after_displace': bool, 'i_cycle_after_escape': bool, 'esc1_modenr_low': int,
        'esc1_modenr_high': int, 'algo_converge': int, 'i_start_after_escape': bool, 'i_cycle_after_converge': bool,
        'algo_escape_again': int, 'i_start_after_converge': bool, 'i_write_initial_eigenvecs': [bool, int]}



# -*- coding: utf-8 -*-
r"""
Base class for spin lattices
"""
from abc import ABC, abstractmethod
import numpy as np
from typing import List, Tuple, Union
from pathlib import Path
import pandas as pd
from spinterface.inputs.lattice.topology.ITopology import ITopology


class ILattice(ABC):
    r"""
    Abstract base class for spin lattices
    """

    def __init__(self, a1: Union[None, np.ndarray], a2: Union[None, np.ndarray], a3: Union[None, np.ndarray],
                 r_motif: Union[None, List[np.ndarray]], N1: Union[None, int], N2: Union[int, None],
                 N3: Union[None, int], magdir: np.ndarray) -> None:
        r"""
        Initializer. ILattice cannot be initialized but the initializer can be called from sub class.

        Args:
            a1(array): lattice vector 1
            a2(array): lattice vector 2
            a3(array): lattice vector 3
            rmotif(List[arrays]): position of the atoms in the unit cell in units of a1, a2, a3
            N1(int): number of unit cells in a1 dir.
            N2(int): number of unit cells in a2 dir.
            N3(int): number of unit cells in a2 dir.
            magdir(array): initial direction of the magnetization
        """
        self.N1 = N1
        self.N2 = N2
        self.N3 = N3
        if r_motif is not None:
            self.NUC = len(r_motif)
        else:
            self.NUC = None
        self.a1 = a1
        self.a2 = a2
        self.a3 = a3
        self._rmotif = r_motif
        if not any([True if element is None else False for element in [N1, N2, N3, r_motif, a1, a2, a3]]):
            self._points, self._magmoms = self._createlattice()
            magdir = magdir / np.linalg.norm(magdir)
            self._spins = np.array([magdir for NN in range(N1 * N2 * N3 * self.NUC)])

    def _createlattice(self) -> Tuple[np.ndarray, np.ndarray]:
        r"""
        Creates the lattice similar to the creation in the spindynamic code.

        Returns:
            the positions of the spins of the lattice in the same order as they appear in some STMi file. Also the mag.
            moments in the same order
        """
        ps, ms = [], []
        for i in range(self.N1):
            for j in range(self.N2):
                for k in range(self.N3):
                    for l in range(self.NUC):
                        p = (i + self._rmotif[l][0]) * self.a1 + (j + self._rmotif[l][1]) * self.a2 + (
                                k + self._rmotif[l][2]) * self.a3
                        ps.append(p)
                        ms.append(self._rmotif[l][3])
        return np.array(ps), np.array(ms)

    def write(self, path: Path = Path.cwd() / 'SpinSTMi.dat') -> None:
        r"""
        Writes the spin lattice to a file
        """
        df = pd.DataFrame(
            {'x': self.X, 'y': self.Y, 'z': self.Z, 'sx': self.SX, 'sy': self.SY, 'sz': self.SZ, 'm': self._magmoms})
        df.to_csv(path, sep=r' ', header=False, index=False)

    @property
    @abstractmethod
    def source(self) -> str:
        r"""
        Returns:
            the type of the lattice (e.g. spin configuration or eigenvector)
        """

    @property
    def spins(self) -> np.ndarray:
        r"""
        Returns:
             the spins in the lattice
        """
        return self._spins

    @spins.setter
    def spins(self, spins: np.ndarray) -> None:
        r"""
        Setter for the magnetisation of the lattice

        Args:
            spins(array): [[sx0, sy0, sz0],...]
        """
        self._spins = spins

    @property
    def points(self) -> np.ndarray:
        r"""
        Returns:
            the points of the lattice.
        """
        return self._points

    @property
    def magstructure(self) -> np.ndarray:
        r"""
        Returns:
            the magnetic structure as used in the spindynamic program
        """
        return np.column_stack((self._points, self._spins, self._magmoms))

    @property
    def X(self) -> np.ndarray:
        r"""
        Returns:
            the X-component of all spins
        """
        return self._points[:, 0]

    @property
    def Y(self) -> np.ndarray:
        r"""
        Returns:
            the Y-component of all spins
        """
        return self._points[:, 1]

    @property
    def Z(self) -> np.ndarray:
        r"""
        Returns:
            the Z-component of all spins
        """
        return self._points[:, 2]

    @property
    def SX(self) -> np.ndarray:
        r"""
        Returns:
            the SX-component of all spins
        """
        return self._spins[:, 0]

    @property
    def SY(self) -> np.ndarray:
        r"""
        Returns:
            the SY-component of all spins
        """
        return self._spins[:, 1]

    @property
    def SZ(self) -> np.ndarray:
        r"""
        Returns:
            the SZ-component of all spins
        """
        return self._spins[:, 2]

    @property
    def nlayer(self) -> int:
        r"""
        Returns:
            the number of layers
        """
        return len(set(self._points[:, 2]))

    @property
    def zlayer(self) -> np.ndarray:
        r"""
        Returns:
            the z coordinates of the occuring layers
        """
        return np.unique(self._points[:, 2])

    @property
    def midpoint(self) -> np.ndarray:
        r"""
        Returns:
            the midpoint of the lattice calculated by the mean
        """
        return np.round(np.mean(self._points, axis=0), 12)

    @property
    def layermidpoints(self) -> np.ndarray:
        r"""
        Returns the midpoints of the layers
        """
        return np.round(np.array([np.mean(self.getlayer_by_idx(l)[:, :3], axis=0) for l in range(self.nlayer)]), 12)

    @property
    def xmagcenter(self) -> np.ndarray:
        r"""
        Returns:
            the center of "mass" (magnetization in x-direction) of the spin lattice
        """
        return 1/self.SX.size * np.sum(self.SX.reshape((np.abs(self.SX.size),1))*self._points, axis=0)

    @property
    def ymagcenter(self) -> np.ndarray:
        r"""
        Returns:
            the center of "mass" (magnetization in x-direction) of the spin lattice
        """
        return 1 / self.SY.size * np.sum(self.SY.reshape((self.SY.size, 1)) * self._points, axis=0)

    @property
    def zmagcenter(self) -> np.ndarray:
        r"""
        Returns:
            the center of "mass" (magnetization in x-direction) of the spin lattice
        """
        return 1 / self.SZ.size * np.sum(self.SZ.reshape((self.SZ.size, 1)) * self._points, axis=0)

    @property
    def total_magnetization(self) -> np.ndarray:
        r"""
        Returns:
            the sum of all spins in the lattice
        """
        return np.sum(self.spins, axis=0)

    def getlayer_by_z(self, z: float) -> np.ndarray:
        r"""
        Args:
            z: z coordinate

        Returns:
            magstructure of the layer with the coordinate z
        """
        return self.magstructure[self.magstructure[:, 2] == z]

    def getlayer_by_idx(self, idx: int) -> np.ndarray:
        r"""
        Args:
            idx: index of the layer
        Returns:
            magstructure of the layer addressed by that index
        """
        if idx > self.nlayer - 1:
            raise ValueError('idx higher than number of layers')
        return self.magstructure[self.magstructure[:, 2] == self.zlayer[idx]]

    def setlayer_by_idx(self, idx: int, magstructure: np.ndarray) -> None:
        r"""
        Sets the magnetisation of a layer

        Args:
            idx(int): index of the layer to set
            magstructure(array): [[x0 y0 z0 sx0 sy0 sz0 mu0],...]
        """
        self._points[self._points[:, 2] == self.zlayer[idx]] = magstructure[:, :3]
        self._spins[self._points[:, 2] == self.zlayer[idx]] = magstructure[:, 3:6]
        self._magmoms[self._points[:, 2] == self.zlayer[idx]] = magstructure[:, 6]

    def setlayer_by_z(self, z: float, magstructure: np.ndarray) -> None:
        r"""
        Sets the magnetisation of a layer

        Args:
            z(foat): z coordinate of the layer to set
            magstructure(array): [[x0 y0 z0 sx0 sy0 sz0 mu0],...]
        """
        self._points[self._points[:, 2] == z] = magstructure[:, :3]
        self._spins[self._points[:, 2] == z] = magstructure[:, 3:6]
        self._magmoms[self._points[:, 2] == z] = magstructure[:, 6]

    @property
    @abstractmethod
    def topologies(self) -> List[ITopology]:
        r"""
        Returns:
            topology information about the lattice
        """





r"""
This classes instances represent topological information of a lattice object.
"""
import numpy as np
import spinterface.inputs.lattice.utilities as utils

from spinterface.inputs.lattice.topology.ITopology import ITopology


class CTopologyHexCubic(ITopology):
    r"""
    Objects of this class hold topology information of spin lattices for hexagonal and cubic in plane geometries
    """

    def __init__(self, points: np.ndarray, spins: np.ndarray, N1: int, N2: int) -> None:
        r"""
        Initializer. The lattice points have to be in one layer.

        Args:
            points: positions of the lattice points of the layer [[p1x,p1y,p1z],...]
            spins: spins of the layer [[s1x,s1y,s1z],...]
            N1, N2: number of spins in N1- and N2-direction
        """
        self.points = points
        self.spins = spins
        self.N1 = N1
        self.N2 = N2
        if not np.all(points[:, 2] == points[0, 2]):
            raise ValueError('ERROR: Not all the points are in the same layer.')
        if any([self.N1 is None, self.N2 is None]):
            raise ValueError('ERROR: N1 or N2 is None. Cant calculate topology')
        self._calculate_topology()

    def _calculate_topology(self) -> None:
        r"""
        Calculate the topology of the given input structure
        """
        spin_indices, dual_lattice_points, areas1, areas2, polygons1, polygons2, topo_dens = [], [], [], [], [], [], []
        for idx_a1 in range(self.N1 - 1):
            for idx_a2 in range(self.N2 - 1):
                # get the indices of the four closest spins
                idx1 = idx_a1 * self.N2 + idx_a2
                idx2 = (idx_a1 + 1) * (self.N2 - 1) + idx_a2 + 1 + idx_a1
                idx3 = (idx_a1 + 1) * (self.N2 - 1) + idx_a2 + 2 + idx_a1
                idx4 = idx_a1 * self.N2 + idx_a2 + 1
                # save the indices for the current dual lattice point
                spin_indices.append(np.array([idx1, idx2, idx3, idx4]))
                # calculate the current point of the dual lattice
                dual_lattice_points.append(
                    np.mean(np.array(
                        [self.points[idx1, :], self.points[idx2, :], self.points[idx3, :], self.points[idx4, :]]),
                        axis=0))
                area1 = utils.spherical_area(s1=self.spins[idx1, :], s2=self.spins[idx2, :], s3=self.spins[idx3])
                area2 = utils.spherical_area(s1=self.spins[idx1, :], s2=self.spins[idx3], s3=self.spins[idx4])
                areas1.append(area1)
                areas2.append(area2)
                polygons1.extend([[self.points[idx1, 0], self.points[idx1, 1], self.points[idx1, 2]],
                                  [self.points[idx2, 0], self.points[idx2, 1], self.points[idx2, 2]],
                                  [self.points[idx3, 0], self.points[idx3, 1], self.points[idx3, 2]]])
                polygons2.extend([[self.points[idx1, 0], self.points[idx1, 1], self.points[idx1, 2]],
                                  [self.points[idx3, 0], self.points[idx3, 1], self.points[idx3, 2]],
                                  [self.points[idx4, 0], self.points[idx4, 1], self.points[idx4, 2]]])
                topo_dens.append((1 / (4 * np.pi)) * (area1 + area2))
        self._spin_indices = np.array(spin_indices)
        self._topo_dens = np.array(topo_dens)
        self._polygons1 = np.array(polygons1)
        self._polygons2 = np.array(polygons2)
        self._areas1 = np.array(areas1)
        self._areas2 = np.array(areas2)
        self._dual_lattice_points = np.array(dual_lattice_points)
        print(f'length topo dens points: {len(topo_dens)}')
        print(f'length dual latt points: {len(dual_lattice_points)}')

    @property
    def dual_lattice_points(self) -> np.ndarray:
        r"""
        Returns:
            the points of the dual lattice
        """
        return self._dual_lattice_points

    @property
    def spin_indices(self) -> np.ndarray:
        r"""
        Returns:
            the indices of the four adjacent spins for the current dual lattice point
        """
        return self._spin_indices

    @property
    def polygons1(self) -> np.ndarray:
        r"""
        Returns:
            the coordinates of the upper triangles for visualizing the topological density
        """
        return self._polygons1

    @property
    def polygons2(self) -> np.ndarray:
        r"""
        Returns:
            the coordinates of the upper triangles for visualizing the topological density
        """
        return self._polygons2

    @property
    def areas1(self) -> np.ndarray:
        r"""
        Returns:
            the spherical area of the triangles associated with polygons 1
        """
        return self._areas1

    @property
    def areas2(self) -> np.ndarray:
        r"""
        Returns:
            the spherical area of the triangles associated with polygons 2
        """
        return self._areas2

    @property
    def topo_dens(self) -> np.ndarray:
        r"""
        Returns:
            the topological density of the dual lattice points
        """
        return self._topo_dens

    @property
    def topological_center(self) -> np.ndarray:
        r"""
        Returns:
            the center of mass with respect to the topological density
        """
        return (1 / self.topologic_charge) * np.sum((self.dual_lattice_points * self.topo_dens[:, None]), axis=0)

    @property
    def topologic_charge(self) -> np.ndarray:
        r"""
        Returns:
            the topologic charge of the lattice
        """
        return np.sum(self.topo_dens)



# -*- coding: utf-8 -*-
r"""
Newton minimization technique for two dimensional functions
"""
import numpy as np
from pathlib import Path
from typing import Callable


class CNewton2D:
    r"""
    Newton minimization technique for 2 dimensional configuration spaces
    """

    def __init__(self, f: Callable, initial: np.ndarray, iterations: int = 20,
                 modification_technique: str = 'nondiagonal_trivial', h: float = 0.00001) -> None:
        r"""
        Initialize newtons method

        Args:
            f(Callable): Callable for energy landscape

            initial(np.ndarray): start point for the calculation

            iterations(int): number of steps

            modification_technique(str): decider string for modification technique of the hessian in non convex regions

            h(float): finite element for hessian approximations
        """
        if len(initial) != 2:
            raise ValueError('Start point has to be 2 dimensional')
        self._initial = initial
        self._f = self._function_wrapper(f=f)
        self._iterations = iterations
        self._modification_technique = modification_technique
        self._h = h

    def _function_wrapper(self, f: Callable) -> Callable:
        r"""
        This wrapper changes the input structure of the callable to be in line with the visualization interface. The
        visualization interface uses a function which takes f(x,y) while the methods in this class need a callable with
        the parameter f(r(x,y)). This wrapper provides this functionality.
        Args:
            f: takes a callable with the functionality f(x,y)

        Returns: a function with f(r).
        """
        def wrapper(r: np.ndarray) -> np.ndarray:
            if len(np.shape(r)) == 1:
                x = r[0]
                y = r[1]
            else:
                x = r[:, 0]
                y = r[:, 1]
            return f(x,y)
        return wrapper

    def __call__(self, i_write_output: bool = False, gralog: int = 1,
                 write_out: Path = Path.cwd() / 'newton.dat') -> np.ndarray:
        r"""
        Calls newtons method

        Args:
            i_write_output(bool): boolean whether an output is written to a file
            gralog(int): frequency for output writing
            write_out(Path): Path for potential output file
        Returns:
            Sequence of point passed by in the minimization. Obtain the minimum through [-1,:]
        """
        points = []
        currentpoint = self._initial
        points.append(currentpoint)
        if i_write_output:
            with open(str(write_out), 'w') as f:
                f.write(f'iter  pointx  pointy  step  forcex  forcey\n')
        for k in range(self._iterations):
            # ensure that hessian is positive definite
            if self._modification_technique == 'nondiagonal_trivial':
                Bk = self._mod_nondiag_trivial(point=currentpoint, delta=1)
            elif self._modification_technique == 'diagonal_trivial':
                Bk = self._mod_diag_trivial(point=currentpoint, delta=1)
            elif self._modification_technique == 'cholesky':
                Bk = self._mod_cholesky(point=currentpoint)
            else:
                raise NotImplementedError('Modification method not implemented!')
            # calculate energy gradient:
            grad = self._gradient(point=currentpoint, h=self._h)
            print('flag0')
            pk = np.linalg.solve(Bk, -1 * grad)
            print('flag1')
            alpha_k = self._choose_steplength(point=currentpoint, step=pk, grad=grad)
            # displace current point
            if np.mod(k, gralog) == 0:
                if i_write_output:
                    with open(str(write_out), 'a') as f:
                        f.write(f'{k}  {currentpoint[0]}  {currentpoint[1]}  {alpha_k}  {pk[0]}  {pk[1]}\n')
                print(f"iteration Newton: {k},point: {currentpoint} , force: {pk * alpha_k}")
            # convergence criterium:
            currentpoint_new = currentpoint + alpha_k * pk
            print('flag2')
            if np.linalg.norm(currentpoint_new - currentpoint) <= 1e-6:
                break
            currentpoint = currentpoint_new
            points.append(currentpoint_new)
        return np.asarray(points)

    def _choose_steplength(self, point: np.ndarray, step: np.ndarray, grad: np.ndarray, reducefactor: float = 0.6,
                           c1: float = 1e-4, startalpha: float = 1) -> float:
        r"""
        Backtracking algorithm with sufficient decrease condition

        Args:
            point(np.ndarray): current point

            step(np.ndarray): current direction of the line search

            grad(np.ndarray): current gradient

            reducefactor(float): reduce factor for alpha. Has to be in interval (0,1)

            c1(float): condition number for sufficient decrease condition. Is usually very small

            startalpha(float): biggest value of alpha which is gradually reduced if condition is not satisfied. For
            Newtons method is should be 1.

        Returns:
            the step length
        """
        alpha = startalpha
        alpha = 1.0
        #while self._f(point + alpha * step) >= (self._f(point) + c1 * alpha * np.dot(grad, step)):
        #    alpha = reducefactor * alpha
        #    print(alpha)
        return alpha

    def _mod_nondiag_trivial(self, point: np.ndarray, delta: float = 1) -> np.ndarray:
        r"""
        Nondiagonal modification technique for delivering a positive definite approx of the hessian.

        Args:
            point(np.ndarray): current point

            delta(float): limit for pos. def. If delta = 1 the negative eigenvalues are mirrored around 1. Another
            option is to set delta larger than machine precision u. E.g. delta= sqrt(u) which is around 10^-8. This
            can leed to really large steps (artificial for Newtons methods).

        Returns:
            positive definite modification of hessian
        """
        # first calculate eigenvalues
        hess = self._hessian(point=point, h=self._h)
        # spectral decomposition the eigenvecs are represented by Q (the columns are the eigenvectors)
        eigenvals, eigenvecs = np.linalg.eigh(hess)
        # build tau_i
        tau = np.array([0.0 if eigenval >= delta else delta - eigenval for eigenval in eigenvals])
        # return modifcated hessian
        return eigenvecs.dot(np.diag(eigenvals + tau)).dot(eigenvecs.T)

    def _mod_diag_trivial(self, point: np.ndarray, delta: float = 1.0) -> np.ndarray:
        r"""
        Diagonal modification technique for delivering a positive definite approx of the hessian.

        Args:
            point(np.ndarray): current point

            delta(float): limit for pos. def. If delta = 1 the negative eigenvalues are mirrored around 1. Another
            option is to set delta larger than machine precision u. E.g. delta= sqrt(u) which is around 10^-8. This
            can leed to really large steps (artificial for Newtons methods).

        Returns:
            positive definite modification of hessian
        """
        # first calculate eigenvalues
        hess = self._hessian(point=point, h=self._h)
        # spectral decomposition the eigenvecs are represented by Q (the columns are the eigenvectors)
        eigenvals, eigenvecs = np.linalg.eigh(hess)
        tau = max(0.0, delta - eigenvals[0])
        return hess + tau * np.identity(2)

    def _mod_cholesky(self, point: np.ndarray, beta: float = 1e-3, iterations: int = 100):
        r"""
        Cholesky with added multiple of identity

        Args:
            point(np.ndarray): current point

            beta(float): heuristic parameter typical value 1e-3

            iterations(int): iterations needed to obtain positive definite matrix

        Returns:
            positive definite approximation of hessian
        """
        # first calculate hessian
        hess = self._hessian(point=point, h=self._h)
        # dimension of problem
        n = hess.shape[0]
        if min([hess[aii][aii] for aii in range(n)]) > 0:
            tau_0 = 0
        else:
            tau_0 = -1 * min([hess[aii][aii] for aii in range(n)]) + beta
        tau_k = tau_0
        for k in range(iterations):
            A = hess + tau_k * np.identity(n)
            try:
                L = np.linalg.cholesky(A)
                return L.dot(L.T)
            except np.linalg.LinAlgError:
                tau_k = max(tau_k * 2, beta)
                continue

    def _hessian(self, point: np.ndarray, h: float) -> np.ndarray:
        r"""
        Calculate the hessian for a given point with finite elements

        Args:
            point(np.ndarray): Point for which the hessian is evaluated.
            h(float): finite element

        Returns:
            numpy array of hessian: [[dfdxdx, dfdxdy],[dfdydx, dfdydy]]
        """
        hess = np.zeros(shape=(2, 2))
        # diagonal terms
        hess[0, 0] = (self._f(point + 2 * h * np.array([1, 0])) - 2 * self._f(point + h * np.array([1, 0])) + self._f(
            point)) / h ** 2
        hess[1, 1] = (self._f(point + 2 * h * np.array([0, 1])) - 2 * self._f(point + h * np.array([0, 1])) + self._f(
            point)) / h ** 2
        hess[0, 1] = (self._f(point + h * np.array([1, 1])) - self._f(point + h * np.array([1, 0])) - self._f(
            point + h * np.array([0, 1])) + self._f(point)) / h ** 2
        hess[1, 0] = hess[0, 1]
        return hess

    def _gradient(self, point: np.ndarray, h: float) -> np.ndarray:
        r"""
        Calculate the energy gradient

        Args:
            point(np.ndarray): Point for which the gradient is evaluated.
            h(float): finite element

        Return:
             1d numpy array with length 2: [dfdx,dfdy]
        """
        dfdx = (self._f(point + h * np.array([1, 0])) - self._f(point)) / h
        dfdy = (self._f(point + h * np.array([0, 1])) - self._f(point)) / h
        return np.array([dfdx, dfdy])



import numpy as np
import pandas as pd
import pyvista as pv
from pathlib import Path
from spinterface.visualizations.energylandscapes.i2dfunction import I2DFunction
from spinterface.visualizations.energylandscapes.cvisualpyvista import CVisualPyVista
from spinterface.algorithms.spineapple.minimizer.cnewton_2d import CNewton2D
from spinterface.algorithms.spineapple.gneb.cneb2d import CNEB2D,CciNEB2d
from spinterface.algorithms.spineapple.modefollowing.cmodefollowing_2d import CModeFollowing2D


class Clandscape(I2DFunction):
    r"""

    """

    def __call__(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        omega_x = 1 / 4 * np.pi  # should produce minimum for x=+-4
        omega_y = 1 / 3 * np.pi  # should produce minimum for y=+-3
        return 1.0 * np.cos(1.0 * (omega_x * x)) + np.cos((omega_y * y)) + \
            self.gauss(x,y,x0=-4, y0=5, a=2.0, sigmax=3.0, sigmay=2) + \
            self.gauss(x,y,x0=-4, y0=-1, a=2.0, sigmay=2, sigmax=3.0)


x = np.linspace(-6, 6.0, 500)
y = np.linspace(0, 6, 500)

func = Clandscape()
visualizer = CVisualPyVista(x=x, y=y, function=func, cam= [(-13.331663532588744, -12.24130749359373, 22.990028902008252),
 (-0.7079845672515099, 2.2874402754346983, 1.3990931175195231),
 (0.3748579064849082, 0.6527011506467436, 0.6583788862730705)])


image = 4

point_size_big = 80 #8ß
point_size_small = 50 #50
# guess points near the minimum based on a view from the top===================================

# start point to minimize into the a state:
newton_minimizer = CNewton2D(f=func, initial=np.array([-4, 1]))
Astate = newton_minimizer()
if image in [1,2,3,4,5]:
    visualizer.placepointonsurface(xypoint=Astate[-1], color='red', point_size=point_size_big)

#visualizer.placepointonsurface(xypoint=np.array([-4, 1]), point_size=point_size_big, color='k')

#.............................................................

#visualizer.placepointonsurface(xypoint=np.array([4, 2]), point_size=point_size_big, color='k')
# start point to minimize into the b state:
newton_minimizer = CNewton2D(f=func, initial=np.array([4, 2]))
Bstate = newton_minimizer()
if image in [1,2,3,4,5]:
    visualizer.placepointonsurface(xypoint=Bstate[-1], color='red', point_size=point_size_big)
#.............................................................
# GNEB A to M
nebber = CNEB2D(f=func,initial=Astate[-1],final=Bstate[-1],nim=15)
#path_AB = nebber(i_write_out=True, write_out=Path.cwd() / 'path_AM.dat')
path_AB = nebber.read_convergedpath_fromfile(Path.cwd() / 'path_AM.dat')
ci_nebber = CciNEB2d(function=func, file=Path.cwd() / 'path_AM.dat')
#path_ABsp, sp = ci_nebber(i_write_out=True,write_out=Path.cwd() / 'path_AB_ci.dat')
path_ABsp = ci_nebber.read_convergedpath_fromfile(Path.cwd() / 'path_AB_ci.dat')
ci = ci_nebber.find_ci()
if image in [1,2,3]:
    visualizer.placepointsequenceonsurface(path_ABsp, color_points='blue', point_size=point_size_small,tuberadius=0.05)
    visualizer.placepointonsurface(xypoint=path_ABsp[ci], color='green', point_size=point_size_big)
#.............................................................

# construct transition state
if image in [2,3]:
    bx = pv.Box(bounds=[0.0,0.1,0.0,6.0,0,3.0])
    visualizer.add_mesh(bx,color='gray',opacity=0.5)

if image in [3]:
    mf = CModeFollowing2D(f=func,initial=Astate[-1])
    # points, evals, evecs = mf(mode_index=0,i_write_out=True,write_out=Path.cwd() / 'mf_0.dat')
    visualizer.create_eigenvalue_parabola(Astate[-1],evals=mf.eigenvalues_initial,evecs=mf.eigenvectors_initial,point_distance_xy=0.1,point_size=30, y_offset=1.5)
    mf = CModeFollowing2D(f=func,initial=path_ABsp[ci])
    # points, evals, evecs = mf(mode_index=0,i_write_out=True,write_out=Path.cwd() / 'mf_0.dat')
    visualizer.create_eigenvalue_parabola(path_ABsp[ci],evals=mf.eigenvalues_initial,evecs=mf.eigenvectors_initial,point_distance_xy=0.1,point_size=30,x_width=2,y_width=2,y_offset=1.5)

if image in [4]:
    mf = CModeFollowing2D(f=func, initial=path_ABsp[ci],mf_steps=20)
    points, evals, evecs = mf(mode_index=0,i_write_out=True,write_out=Path.cwd() / 'mode1.dat')
    visualizer.placepointsequenceonsurface(seq_xypoint=np.asarray(points),color_points='tab:orange')
    points, evals, evecs = mf(mode_index=1,i_write_out=True,write_out=Path.cwd() / 'mode2.dat')
    visualizer.placepointsequenceonsurface(seq_xypoint=np.asarray(points),color_points='tab:orange')


visualizer.show()
visualizer(outpath=Path.cwd() / f'method_1sp{image}.png')



r"""
Base class for visualizing a certain stage of the results of the application of the sps-algorithm
"""
from abc import ABC, abstractmethod
from typing import List, Union, Tuple


class IVisStage(ABC):
    r"""
    Abstract base class for visualizing a certain stage
    """

    def __init__(self, strategy: int, camera_spin_files: Union[List[Tuple[float, float, float]], None] = None) -> None:
        r"""
        Initialize the visualization stage
        Args:
            strategy(int): for each stage a certain strategy was applied. The visualization also depends on these strategies.
        """
        self.strategy = strategy
        self.spincam = camera_spin_files

    @property
    @abstractmethod
    def stage(self) -> str:
        r"""
        Returns:
            the name of the stage
        """

    @abstractmethod
    def _stagestrategytext(self) -> str:
        r"""
        Returns:
            the strategy of the stage
        """

    def __repr__(self) -> None:
        r"""
        Represents the strategy for the stage in words.
        """
        print(f'Visualizer Stage: {self.stage}.')
        print(f'Strategy: {self._stagestrategytext()}')

    @abstractmethod
    def __call__(self) -> None:
        r"""
        Calls the certain visualization.
        """
